// Fake language.dll to inject IPv6 code into Tiberian Sun.
// Copyright 2010 Steinar H. Gunderson <sgunderson@bigfoot.com>.
//
// Tested with mingw32 only; may or may not compile with MSVC.
// To compile, (assuming cross-compile from Linux :-) ):
//
//   i586-mingw32msvc-gcc -Wall -shared -o language.dll ts-ipv6-dll.c -lws2_32         
//
// See http://ts.sesse.net/ for more information.

#include <stdio.h>
#include <ws2tcpip.h>

#define IPV6_V6ONLY 27
#define LOG_TO_FILE

// thread safety problems ahoy
int inited = 0;
HMODULE language_orig, language_this;

#ifdef LOG_TO_FILE
#define LOG_FP log_fp
FILE *log_fp = NULL;
#else
#define LOG_FP stderr
#endif

#define MAX_TRANSLATION_ENTRIES 4096
struct translation_entry {
	struct in_addr addr4;
	struct in6_addr addr6;
};
struct translation_entry translation_entries[MAX_TRANSLATION_ENTRIES];
int num_translation_entries = 0;

void __cdecl output_debug(const char *fmt, ...) {
	va_list ap, aq;
	va_start(ap, fmt);
	va_copy(aq, ap);
	va_end(ap);
	vfprintf(LOG_FP, fmt, aq);
}

struct hostent* __stdcall my_gethostbyname(const char *name)
{
	fprintf(LOG_FP, "gethostbyname(%s)\n", name);
	return gethostbyname(name);
}

int __stdcall my_socket(int af, int type, int protocol)
{
	int sock;
	if (af == AF_INET) {
#if 0
		fprintf(LOG_FP, "Creating socket\n");
		sock = socket(af, type, protocol);
#else
		fprintf(LOG_FP, "Creating IPv6 socket (type=%d, protocol=%d)", type, protocol);
		sock = socket(AF_INET6, type, protocol);
		fprintf(LOG_FP, " = %d\n", sock);
		if (sock != -1) {
			int zero = 0;
			int ret = setsockopt(sock, IPPROTO_IPV6, IPV6_V6ONLY, (char *)&zero, sizeof(zero));
			fprintf(LOG_FP, "IPV6_V6ONLY=0 returned %d (%d)\n", ret, WSAGetLastError());
		}
#endif
	} else {
		fprintf(LOG_FP, "Creating socket of unknown type\n");
		sock = socket(af, type, protocol);
	}
	return sock;
}

void force_addr6(const struct sockaddr *addr, struct sockaddr_in6 *addr6)
{
	switch (addr->sa_family) {
	case AF_INET: {
		int i;
		const struct sockaddr_in *addr4 = (const struct sockaddr_in *)addr;
		memset(addr6, 0, sizeof(*addr6));
		addr6->sin6_family = AF_INET6;
		addr6->sin6_port = addr4->sin_port;

		for (i = 0; i < num_translation_entries; ++i) {
			if (memcmp(&addr4->sin_addr, &translation_entries[i].addr4, sizeof(addr4->sin_addr)) == 0) {
				addr6->sin6_addr = translation_entries[i].addr6;
				return;
			}
		}

		addr6->sin6_addr.s6_addr32[0] = 0;
		addr6->sin6_addr.s6_addr32[1] = 0;
		addr6->sin6_addr.s6_addr32[2] = htonl(0xffff);
		addr6->sin6_addr.s6_addr32[3] = addr4->sin_addr.s_addr;
		break;
	}
	case AF_INET6:
		memcpy(addr6, addr, sizeof(*addr6));
		break;
	default:
		fprintf(LOG_FP, "Trying to force unknown address family %d into IPv6 socket\n", addr->sa_family);
		memcpy(addr6, addr, sizeof(struct sockaddr));
	}
}

int __stdcall my_connect(int sockfd, struct sockaddr *addr, int addrlen)
{
	struct sockaddr_in6 addr6;
	char addrstr[256];
	DWORD addrstr_len = sizeof(addrstr);
	int ret;

	force_addr6(addr, &addr6);

	// Connect to the (IPv6-capable) test server instead of the main server.
	if (addr6.sin6_port == htons(4001)) {
		addr6.sin6_port = htons(4000);
	}

	WSAAddressToStringA((struct sockaddr *)&addr6, sizeof(addr6), NULL, addrstr, &addrstr_len);
	fprintf(LOG_FP, "Connecting socket %d to %s", sockfd, addrstr);
	ret = connect(sockfd, (struct sockaddr *)&addr6, sizeof(addr6));

	if (ret == -1) {
		fprintf(LOG_FP, " = %d (%d)\n", ret, WSAGetLastError());
	} else {
		fprintf(LOG_FP, " = %d\n", ret);
	}
	return ret;
}

void normalize_address(const struct sockaddr *src, struct sockaddr_storage *dst, int *dst_len)
{
	switch (src->sa_family) {
	case AF_INET:
		memcpy(dst, src, sizeof(struct sockaddr_in));
		*dst_len = sizeof(struct sockaddr_in);
		break;
	case AF_INET6: {
		int i;
		const struct sockaddr_in6 *src6 = (const struct sockaddr_in6 *)src;
		if (src6->sin6_addr.s6_addr32[0] == 0 &&
		    src6->sin6_addr.s6_addr32[1] == 0 &&
		    src6->sin6_addr.s6_addr32[2] == htonl(0xffff)) {
			struct sockaddr_in *dst4 = (struct sockaddr_in *)dst;
			dst4->sin_family = AF_INET;
			dst4->sin_addr.s_addr = src6->sin6_addr.s6_addr32[3];
			dst4->sin_port = src6->sin6_port;
			*dst_len = sizeof(struct sockaddr_in);
			return;
		}
		for (i = 0; i < num_translation_entries; ++i) {
			if (memcmp(&src6->sin6_addr, &translation_entries[i].addr6, sizeof(src6->sin6_addr)) == 0) {
				struct sockaddr_in *dst4 = (struct sockaddr_in *)dst;
				dst4->sin_family = AF_INET;
				dst4->sin_addr = translation_entries[i].addr4;
				dst4->sin_port = src6->sin6_port;
				*dst_len = sizeof(struct sockaddr_in);
				return;
			}
		}
		memcpy(dst, src, sizeof(struct sockaddr_in6));
		*dst_len = sizeof(struct sockaddr_in6);
		break;
	}
	default:
		fprintf(LOG_FP, "Trying to normalize unknown address family %d\n", src->sa_family);
		memcpy(dst, src, sizeof(struct sockaddr));
		*dst_len = sizeof(struct sockaddr);
	}
}

int __stdcall my_recvfrom(int sock, char* buf, int len, int flags, struct sockaddr* from, int *fromlen)
{
	struct sockaddr_storage real_from;
	struct sockaddr_storage normalized_from;
	int real_fromlen = sizeof(real_from);
	int normalized_len;
	char addrstr[256];
	DWORD addrstr_len = sizeof(addrstr);
	int ret;

	ret = recvfrom(sock, buf, len, flags, (struct sockaddr *)&real_from, &real_fromlen);
	if (ret >= 0) {
		normalize_address((struct sockaddr *)&real_from, &normalized_from, &normalized_len);
		if (normalized_len > *fromlen) {
			memcpy(from, &normalized_from, *fromlen);
		} else {
			memcpy(from, &normalized_from, normalized_len);
		}
		*fromlen = normalized_len;

		WSAAddressToStringA((struct sockaddr *)&normalized_from, normalized_len, NULL, addrstr, &addrstr_len);
		fprintf(LOG_FP, "Received %d bytes from %s (family=%d, sockfd=%d)\n", ret, addrstr, normalized_from.ss_family, sock);
	} else {
		fprintf(LOG_FP, "recv(sockfd=%d) returned %d\n", sock, ret);
	}

	return ret;
}

ssize_t __stdcall my_sendto(int sockfd, void *buf, size_t len, int flags, struct sockaddr *dest_addr, socklen_t addrlen)
{
	char addrstr[256];
	DWORD addrstr_len = sizeof(addrstr);
	ssize_t ret;
	struct sockaddr_in6 addr6;
	
	force_addr6(dest_addr, &addr6);
	WSAAddressToStringA((struct sockaddr *)&addr6, sizeof(addr6), NULL, addrstr, &addrstr_len);

	fprintf(LOG_FP, "Sending %d bytes to %s (family=%d, socket=%d)", len, addrstr, addr6.sin6_family, sockfd);
	ret = sendto(sockfd, buf, len, flags, (struct sockaddr *)&addr6, sizeof(addr6));
	fprintf(LOG_FP, " = %ld (%d)\n", ret, WSAGetLastError());
	return ret;
}

int __stdcall my_bind(int sockfd, const struct sockaddr *addr, socklen_t addrlen)
{
	char addrstr[256];
	DWORD addrstr_len = sizeof(addrstr);
	ssize_t ret;
	struct sockaddr_in6 addr6;
	
	force_addr6(addr, &addr6);
	WSAAddressToStringA((struct sockaddr *)&addr6, sizeof(addr6), NULL, addrstr, &addrstr_len);

	fprintf(LOG_FP, "bind(sockfd=%d, addr=%s)", sockfd, addrstr);
	ret = bind(sockfd, (struct sockaddr *)&addr6, sizeof(addr6));
	fprintf(LOG_FP, " = %ld (%d)\n", ret, WSAGetLastError());
	return ret;
}

int __stdcall my_getpeername(int sockfd, struct sockaddr *addr, socklen_t *addrlen)
{
	struct sockaddr_storage real_from;
	struct sockaddr_storage normalized_from;
	int real_fromlen = sizeof(real_from);
	int normalized_len;
	char addrstr[256];
	DWORD addrstr_len = sizeof(addrstr);
	int ret;

	fprintf(LOG_FP, "getpeername(sockfd=%d)", sockfd);
	ret = getpeername(sockfd, (struct sockaddr *)&real_from, &real_fromlen);
	if (ret >= 0) {
		normalize_address((struct sockaddr *)&real_from, &normalized_from, &normalized_len);
		if (normalized_len > *addrlen) {
			memcpy(addr, &normalized_from, *addrlen);
		} else {
			memcpy(addr, &normalized_from, normalized_len);
		}
		*addrlen = normalized_len;

		WSAAddressToStringA((struct sockaddr *)&normalized_from, normalized_len, NULL, addrstr, &addrstr_len);
		fprintf(LOG_FP, " = %s\n", addrstr);
	} else {
		fprintf(LOG_FP, " = %d (%d)\n", ret, WSAGetLastError());
	}

	return ret;
}

unsigned lookup_natpt_entry(const char *str)
{
	struct sockaddr_in6 addr6;
	struct in_addr addr4;	
	int len = sizeof(addr6);
	int ret, i;
	char *dupstr = strdup(str);
	char *ptr;

	ptr = strchr(dupstr, '\r');
	if (ptr != NULL) {
		*ptr = 0;
	}

	ptr = strchr(dupstr, '\n');
	if (ptr != NULL) {
		*ptr = 0;
	}

	ptr = strchr(dupstr, '_');
	while (ptr != NULL) {
		*ptr = ':';
		ptr = strchr(ptr, '_');
	}

	ret = WSAStringToAddress(dupstr, AF_INET6, NULL, (struct sockaddr *)&addr6, &len);
	free(dupstr);
	if (ret != 0) {
		fprintf(LOG_FP, "Malformed IPv6 address: '%s'\n", str);
		return atol(str);
	}

	for (i = 0; i < num_translation_entries; ++i) {
		if (memcmp(&addr6.sin6_addr, &translation_entries[i].addr6, sizeof(addr6.sin6_addr)) == 0) {
			fprintf(LOG_FP, "Parsing '%s' to %ld\n", str, translation_entries[i].addr4.s_addr);
			return ntohl(translation_entries[i].addr4.s_addr);
		}
	}

	addr4.s_addr = htonl((224 << 24) | (num_translation_entries + 1));
	fprintf(LOG_FP, "New auto-NAT-PT entry: %s -> %s\n", str, inet_ntoa(addr4));

	translation_entries[num_translation_entries].addr4 = addr4;
	translation_entries[num_translation_entries].addr6 = addr6.sin6_addr;
	++num_translation_entries;

	return ntohl(addr4.s_addr);
}
	
int __cdecl my_ipv6_atol(const char *str)
{
	if (strchr(str, '_') == NULL) {
		return atol(str);
	}
	return lookup_natpt_entry(str);
}

int __cdecl other_atol(const char *str)
{
	int myval = 0x1234;
//	int i;
	fprintf(LOG_FP, "atol('%s'), estimated return address = %p\n", str, *(int**)(&myval + 2));
#if 0
	fprintf(LOG_FP, "Stack:");
	for (i = 0; i < 20; ++i) {
		fprintf(LOG_FP, " %p", *(int**)(&myval + i));
	}
	fprintf(LOG_FP, "\n");
#endif
	return atol(str);
}

void func_replace_hotpatch(DWORD address, void* new_address)
{
	unsigned char jmp_opcode = 0xe9;
	DWORD relative_addr = (DWORD)new_address - (DWORD)address - 5;

	VirtualProtect((void*)address, 5, PAGE_EXECUTE_READWRITE, NULL);
	memcpy((void*)address, &jmp_opcode, 1);
	memcpy((void*)address + 1, &relative_addr, sizeof(DWORD));
}
	
void func_replace_offset(DWORD address, void* new_address)
{
	VirtualProtect((void*)address, 4, PAGE_EXECUTE_READWRITE, NULL);
	memcpy((void*)address, &new_address, sizeof(DWORD));
}

void func_replace_relative_call(DWORD address, void* new_address)
{
	DWORD relative_addr = (DWORD)new_address - (address + 5);
	VirtualProtect((void*)address, 5, PAGE_EXECUTE_READWRITE, NULL);
	memcpy((char*)address + 1, &relative_addr, sizeof(DWORD));
}

void init_winsock()
{
	WORD wVersionRequested;
	WSADATA wsaData;
	wVersionRequested = MAKEWORD(2, 2);
	int err = WSAStartup(wVersionRequested, &wsaData);
	fprintf(LOG_FP, "Winsock inited, err=%d\n", err);
}

#if 0
void read_translation_file()
{
	FILE *fp;
	char buf[256];

	fp = fopen("natpt.txt", "r");
	if (fp == NULL) {
		perror("natpt.txt");
		fprintf(LOG_FP, "Couldn't open natpt.txt, not doing NAT-PT.\n");
		return;
	}

	num_translation_entries = 0;
	while (1) {
		char *ptr;
		int ret, len;
		struct sockaddr_in addr4;
		struct sockaddr_in6 addr6;

		if (fgets(buf, 256, fp) == NULL) {
			fprintf(LOG_FP, "natpt.txt: Read %d translation entries.\n", num_translation_entries);
			fclose(fp);
			return;
		}
		if (num_translation_entries >= MAX_TRANSLATION_ENTRIES) {
			fprintf(LOG_FP, "natpt.txt: Only room for %d translation entries, truncating before end.\n", num_translation_entries);
			fclose(fp);
			return;
		}
		ptr = strchr(buf, '\n');
		if (ptr != NULL) {
			*ptr = 0;
		}
		ptr = strchr(buf, '\r');
		if (ptr != NULL) {
			*ptr = 0;
		}
		ptr = strchr(buf, ' ');
		if (ptr == NULL) {
			fprintf(LOG_FP, "natpt.txt: Malformed line: '%s'\n", buf);
			continue;
		}

		*ptr = 0;
		len = sizeof(addr4);
		ret = WSAStringToAddress(buf, AF_INET, NULL, (struct sockaddr *)&addr4, &len);
		if (ret != 0) {
			fprintf(LOG_FP, "natpt.txt: Malformed IPv4 address: '%s'\n", buf);
			continue;
		}
		translation_entries[num_translation_entries].addr4 = addr4.sin_addr;
	
		len = sizeof(addr6);	
		ret = WSAStringToAddress(ptr + 1, AF_INET6, NULL, (struct sockaddr *)&addr6, &len);
		if (ret != 0) {
			fprintf(LOG_FP, "natpt.txt: Malformed IPv6 address: '%s'\n", ptr + 1);
			continue;
		}
		translation_entries[num_translation_entries].addr6 = addr6.sin6_addr;
		++num_translation_entries;
	}
}
#else
void read_translation_file()
{
	int ret, len;
	struct sockaddr_in6 addr6;

	struct sockaddr_in addr4;
	len = sizeof(addr4);
	ret = WSAStringToAddress("0.0.0.0", AF_INET, NULL, (struct sockaddr *)&addr4, &len);
	if (ret != 0) {
		fprintf(LOG_FP, "Error parsing address 0.0.0.0 (ret=%d, err=%d, len=%d)\n", ret, WSAGetLastError(), len);
		return;
	}
	translation_entries[num_translation_entries].addr4 = addr4.sin_addr;
	
	len = sizeof(addr6);	
	ret = WSAStringToAddress("::", AF_INET6, NULL, (struct sockaddr *)&addr6, &len);
	if (ret != 0) {
		fprintf(LOG_FP, "Error parsing address ::\n");
		return;
	}
	
	num_translation_entries = 1;
}
#endif

int __stdcall my_LoadStringA(HINSTANCE hInstance, UINT uID, LPSTR lpBuffer, int nBufferMax)
{
	int ret;
	if (hInstance == language_this) {
		hInstance = language_orig;
	}
	ret = LoadStringA(hInstance, uID, lpBuffer, nBufferMax);
	if (hInstance == language_orig && uID == 1) {
		strcat(lpBuffer, " / IPv6 patch \xa9 2010 Steinar H. Gunderson");
	}
	return ret;
}

HRSRC __stdcall my_FindResourceA(HMODULE hModule, LPCTSTR lpName, LPCTSTR lpType)
{
	if (hModule == language_this) {
		hModule = language_orig;
	}
	return FindResourceA(hModule, lpName, lpType);
}

HGLOBAL __stdcall my_LoadResource(HMODULE hModule, HRSRC hResInfo)
{
	if (hModule == language_this) {
		hModule = language_orig;
	}
	return LoadResource(hModule, hResInfo);
}

unsigned char initial_conversion_code[] = {
	0xe8, 0x00, 0x00, 0x00, 0x00,         // call imm32
	0x5f,                                 // pop edi
	0x58,                                 // pop eax
	0x68, 0x00, 0x00, 0x00, 0x00,         // push imm32
	0x57,                                 // push edi
	0xff, 0x15, 0x08, 0xe2, 0x02, 0x21,   // call wsprintfA
	0xe9, 0x00, 0x00, 0x00, 0x00,         // jmp imm32
};
char new_initial_conversion[256];
	
void __stdcall update_conversation()
{
	int sock, ret, len;
	struct sockaddr_in6 dest, src;
	char addrstr[256], extraconv[256], *ptr;
	DWORD addrstr_len = sizeof(addrstr);
	strcpy(new_initial_conversion, (char *)0x21032f30);  // old conversation

	// Use the standard trick of connect()-ing an UDP socket to some globally
	// known destination to get our local IPv6 address. We won't get the best
	// address this way (among others, it uses a temporary address when privacy
	// extensions are in effect), but it sure is convenient. :-)
	sock = socket(AF_INET6, SOCK_DGRAM, IPPROTO_UDP);
	if (sock == -1) {
		fprintf(LOG_FP, "sock fail\n");
		return;
	}
		
	len = sizeof(dest);
	ret = WSAStringToAddress("2001:700:300:1880::2", AF_INET6, NULL, (struct sockaddr *)&dest, &len);
	if (ret != 0) {
		fprintf(LOG_FP, "addr1 fail\n");
		closesocket(sock);
		return;
	}

	ret = connect(sock, (struct sockaddr *)&dest, len);

	len = sizeof(src);	
	ret = getsockname(sock, (struct sockaddr *)&src, &len);
	closesocket(sock);
	if (ret == -1) {
		fprintf(LOG_FP, "getsockname fail\n");
		return;
	}

	src.sin6_port = 0;
	ret = WSAAddressToStringA((struct sockaddr *)&src, sizeof(src), NULL, addrstr, &addrstr_len);
	if (ret == -1) {
		fprintf(LOG_FP, "wsaaddresstostring fail (%d)\n", WSAGetLastError());
		return;
	}
	
	ptr = strchr(addrstr, ':');
	while (ptr) {
		*ptr = '_';
		ptr = strchr(ptr, ':');
	}

	sprintf(extraconv, "IPV6ADDR %s\r\n", addrstr);	
	strcat(new_initial_conversion, extraconv);

	fprintf(LOG_FP, "Updated conversion to: '%s'\n", new_initial_conversion);
}

void replace_initial_conversation()
{
	// fill in the address to the function we want to call
	char *funcptr = (char *)update_conversation - ((DWORD)initial_conversion_code + 5);
	memcpy(initial_conversion_code + 1, &funcptr, sizeof(funcptr));

	// change the conversation variable
	char *strptr = new_initial_conversion;
	memcpy(initial_conversion_code + 8, &strptr, sizeof(strptr));

	// fill in jump back
	DWORD relative_addr_back = 0x21008282 - ((DWORD)initial_conversion_code + sizeof(initial_conversion_code));
	memcpy(initial_conversion_code + sizeof(initial_conversion_code) - sizeof(DWORD), &relative_addr_back, sizeof(DWORD));

	// patch in new code
	VirtualProtect((void*)0x2100827c, 5, PAGE_EXECUTE_READWRITE, NULL);
	VirtualProtect((void*)initial_conversion_code, sizeof(initial_conversion_code), PAGE_EXECUTE_READWRITE, NULL);
	unsigned char jmp_opcode = 0xe9;
	DWORD relative_addr = (DWORD)initial_conversion_code - (0x2100827c + 5);
	memcpy((void*)0x2100827c, &jmp_opcode, 1);
	memcpy((void*)0x2100827d, &relative_addr, sizeof(DWORD));
}
	
void init()
{
#ifdef LOG_TO_FILE
	log_fp = fopen("tsipv6.txt", "w");
#endif
	init_winsock();
	read_translation_file();

	language_orig = LoadLibrary("languageorig.dll");

	fprintf(LOG_FP, "Redirecting resource lookups.\n");
	func_replace_offset(0x6ca45c, my_LoadStringA);
	func_replace_offset(0x6ca204, my_FindResourceA);
	func_replace_offset(0x6ca208, my_LoadResource);

#if 0
	fprintf(LOG_FP, "Enabling debug output.\n");
	func_replace_hotpatch(0x4082d0, output_debug);
#endif
	
	fprintf(LOG_FP, "Hot-patching in IPv6 shim.\n");
	func_replace_offset(0x6ca534, my_gethostbyname);
	func_replace_offset(0x6ca53c, my_socket);
	func_replace_offset(0x6ca540, my_bind);
	func_replace_offset(0x6ca4fc, my_recvfrom);
	func_replace_offset(0x6ca504, my_sendto);

	LoadLibraryA("wolapi.dll");
	func_replace_offset(0x2102e2a4, my_socket);
	func_replace_offset(0x2102e26c, my_connect);
	func_replace_offset(0x2102e28c, my_getpeername);
	
	fprintf(LOG_FP, "Hot-patching in new protocol code.\n");
	replace_initial_conversation();
	func_replace_relative_call(0x2100afcc, my_ipv6_atol);  // 353 parsing?
	func_replace_relative_call(0x2100abbe, my_ipv6_atol);  // JOINGAME parsing
	func_replace_relative_call(0x2100b55f, my_ipv6_atol);  // JOINGAME parsing???
	//func_replace_hotpatch(0x2101cf18, other_atol);
}

void ensure_init()
{
	if (!inited) {
		init();
		inited = 1;
	}
}

int __stdcall DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	language_this = hinstDLL;
	ensure_init();
	return TRUE;
}
